#define bool unsigned char
	#define true  1
	#define false 0

#define byte unsigned char

#define LPC_LOGICAL_DEVICE_OFFSET 0x07

#include <dos.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

unsigned short EFER, EFIR, EFDR;

void ShowUsage(void)
{
	printf("\n-----------------------------------------------------");
	printf("\n Invalid setting, Unit should be 0 or 1.");
	printf("\n Usage: WDT Val Unit, unit=0 (second); unit=1 (min)");
	printf("\n Ex: WDT 10 0, watchdog (pin 77) assert after 10 sec");
	printf("\n Ex: WDT 5 1, watchdog assert after 5 min");
	printf("\n-----------------------------------------------------");
}

bool _EnableWDTO(byte val, byte unit)
{
	byte bData = 0xFF;

	// 2Ah, Bit[3] = 0b
	outportb(EFIR, 0x2A);
	bData = inportb(EFDR);
	bData = (byte)(bData & 0xF7);
	outportb(EFDR, bData);

	// (WDT_EN) 2Dh, Bit[0] = 0b
	outportb(EFIR, 0x2D);
	bData = inportb(EFDR);
	bData = (byte)(bData & 0xFE);
	outportb(EFDR, bData);

	// WDT1
	outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
	outportb(EFDR, 0x08);

	outportb(EFIR, 0x30);
	bData = inportb(EFDR);
	bData = (byte)(bData | 0x01);
	outportb(EFDR, bData);

	outportb(EFIR, 0xF0);
	bData = inportb(EFDR);
	bData = (byte)(bData & 0xE0);
	// bData = (byte)(bData | 0x02); // use KBRST# to reset system

	if (unit == 1) bData = (byte)(bData | 0x08); // unit = min
	outportb(EFDR, bData);

	outportb(EFIR, 0xF1);
	outportb(EFDR, (byte)val);

	outportb(EFIR, 0xF2);
	bData = inportb(EFDR);
	bData = (byte)(bData & 0xE0);
	if (unit == 1) bData = (byte)(bData & 0xEF); // clear status bit
	outportb(EFDR, bData);

	return true;
}


int main(int argc, char **argv)
{
	bool bRet, IsDeviceReady;
	unsigned short i, j, Hi, Lo;
	byte ucTemp, ucHi, ucLo, Val, Unit;

	// process command line arguments
	//
	// WDT 10 0, watchdog assert after 10 sec
	// WDT 5 1, watchdog assert after 5 min
	//

	for (i = 1; i < argc; i++)
	{
		if (strlen(argv[i]) > 30)
		{
			printf("\nInvalid Argument, %s.", argv[i]);

			ShowUsage();
			return 1;
		}

		// force to translate to capital
		// strupr(argv[i]);
		j = atoi(argv[i]);
		if (i == 1) Val = j;
		if (i == 2) Unit = j;
	}

	if ((Unit != 0) && (Unit != 1))
	{
		ShowUsage();
		return 1;
	}

	IsDeviceReady = false;
	for (i = 0; i < 2; i++)
	{
		if (i == 0) EFER = 0x4E;
		else        EFER = 0x2E;

		EFIR = EFER;
		EFDR = (unsigned short)(EFER + 1);
		outportb(EFER, 0xAA);
		outportb(EFER, 0x87);
		outportb(EFER, 0x87);

		outportb(EFIR, 0x20);
		ucHi = inportb(EFDR);

		outportb(EFIR, 0x21);
		ucLo = inportb(EFDR);

		if ((ucHi != 0xFF) || (ucLo != 0xFF))
		{
			if ((ucHi == 0xD2) && ((ucLo & 0x80) == 0x80))
			{
				// CHIP = NCT6116;
				IsDeviceReady = true;
				break;
			}
		}
	}

	if (IsDeviceReady == false)
	{
		printf("\n NCT6116 is not found !");
		return 1;
	}

	// Enable pin 77, WDTO
	if (Unit == 1) _EnableWDTO(Val, 1);
	else           _EnableWDTO(Val, 0);

	return 0;
}



