﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.IO;
using System.Threading;
using System.Management;
using System.Reflection;
//using log4net;
//using log4net.Config;
using NNuvHwAccCtrl;

namespace NNuvHwAccCtrl
{
    public class DriverManager
    {
        #region Constant
        private const string NUV_DRV_SERVICE_NAME = "WBHWDOCT";
        private const string NUV_DRV_DISPLAY_NAME = "Nuvoton Hardware Monitor Driver";
        private const string NUV_DRV_DEVICE_PATH = "\\\\.\\WBHWDOCTDev";
        private const string NUV_DRV_BINARY_X86 = "WBHWDOCT.SYS";
        private const string NUV_DRV_BINARY_X64 = "WBHWDOCT64.SYS";
        private const string NUV_DRV_DIRECTORY = "Drivers";
        
        /// <summary>
        /// constants lifted from winioctl.h from platform sdk
        /// </summary>
        private const uint GPD_TYPE = 40000;
        private const uint METHOD_NEITHER = 3;
        private const uint METHOD_BUFFERED = 0;
        private const uint FILE_ANY_ACCESS = 0;
        private const uint FILE_SPECIAL_ACCESS = FILE_ANY_ACCESS;
        private const uint FILE_READ_ACCESS = 1;
        private const uint FILE_WRITE_ACCESS = 2;

        private const uint GPD_IOCTL_READ_PORT_BASE_INDEX = 0x900;
        private const uint GPD_IOCTL_WRITE_PORT_BASE_INDEX = 0x910;
        private const uint GPD_IOCTL_OTHER_OPERATION_BASE_INDEX = 0x902;
        private const uint GPD_IOCTL_READ_PHY_ADD_BASE_INDEX = 0x930;
        private const uint GPD_IOCTL_WRITE_PHY_ADD_BASE_INDEX = 0x940;

        private static uint IOCTL_GPD_READ_PORT_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PORT_BASE_INDEX + 0, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_READ_PORT_USHORT = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PORT_BASE_INDEX + 1, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_READ_PORT_ULONG = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PORT_BASE_INDEX + 2, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_READ_PORT_BUFFER_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PORT_BASE_INDEX + 3, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_WRITE_PORT_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PORT_BASE_INDEX + 0, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        private static uint IOCTL_GPD_WRITE_PORT_USHORT = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PORT_BASE_INDEX + 1, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        private static uint IOCTL_GPD_WRITE_PORT_ULONG = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PORT_BASE_INDEX + 2, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        private static uint IOCTL_GPD_WRITE_PORT_BUFFER_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PORT_BASE_INDEX + 3, METHOD_BUFFERED, FILE_WRITE_ACCESS);

        private static uint IOCTL_GPD_READ_PHY_MEM_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PHY_ADD_BASE_INDEX + 0, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_READ_PHY_MEM_USHORT = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PHY_ADD_BASE_INDEX + 1, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_READ_PHY_MEM_ULONG = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PHY_ADD_BASE_INDEX + 2, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_READ_PHY_MEM_ULONG64 = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PHY_ADD_BASE_INDEX + 3, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_READ_PHY_MEM_BUFFER_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_READ_PHY_ADD_BASE_INDEX + 4, METHOD_BUFFERED, FILE_READ_ACCESS);
        private static uint IOCTL_GPD_WRITE_PHY_MEM_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PHY_ADD_BASE_INDEX + 0, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        private static uint IOCTL_GPD_WRITE_PHY_MEM_USHORT = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PHY_ADD_BASE_INDEX + 1, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        private static uint IOCTL_GPD_WRITE_PHY_MEM_ULONG = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PHY_ADD_BASE_INDEX + 2, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        private static uint IOCTL_GPD_WRITE_PHY_MEM_ULONG64 = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PHY_ADD_BASE_INDEX + 3, METHOD_BUFFERED, FILE_WRITE_ACCESS);
        private static uint IOCTL_GPD_WRITE_PHY_MEM_BUFFER_UCHAR = CTL_CODE(GPD_TYPE, GPD_IOCTL_WRITE_PHY_ADD_BASE_INDEX + 4, METHOD_BUFFERED, FILE_WRITE_ACCESS);

        public static uint CTL_CODE(uint DeviceType, uint Function, uint Method, uint Access)
        {
            return ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method);
        }

        [DllImport("kernel32.dll")]
        private static extern void GetNativeSystemInfo(out SYSTEM_INFO lpSystemInfo);

        [DllImport("kernel32.dll")]
        private static extern bool GetVersionEx(ref OSVERSIONINFOEX osVersionInfo);

        [DllImport("kernel32.dll")]
        private static extern bool IsWow64Process(IntPtr hProcess, out bool lpSystemInfo);

        // wProcessorArchitecture field of SYSTEM_INFO Structure
        private const ushort PROCESSOR_ARCHITECTURE_INTEL = 0;
        private const ushort PROCESSOR_ARCHITECTURE_AMD64 = 9;
        private const ushort PROCESSOR_ARCHITECTURE_IA64 = 6;
        private const ushort PROCESSOR_ARCHITECTURE_UNKNOWN = 0xFFFF;

        // dwPlatformId field of OSVERSIONINFOEX Structure
        private const uint VER_PLATFORM_WIN32_WINDOWS = 1;
        private const uint VER_PLATFORM_WIN32_NT = 2;

        // wProductType field of OSVERSIONINFOEX Structure
        private const ushort VER_NT_SERVER = 3;
        private const ushort VER_NT_DOMAIN_CONTROLLER = 2;
        private const ushort VER_NT_WORKSTATION = 1;
        #endregion

        #region Structure
        [StructLayout(LayoutKind.Explicit)]
        private struct GENPORT_WRITE_INPUT
        {
            [FieldOffset(0)]
            public UInt32 PortNumber;
            [FieldOffset(4)]
            public byte CharData;
            [FieldOffset(4)]
            public ushort ShortData;
            [FieldOffset(4)]
            public UInt32 LongData;
        }

        [StructLayout(LayoutKind.Explicit)]
        private struct GENMEM_WRITE_INPUT
        {
            [FieldOffset(0)]
            public UInt64 MemAddr;
            [FieldOffset(8)]
            public byte CharData;
            [FieldOffset(8)]
            public UInt16 ShortData;
            [FieldOffset(8)]
            public UInt32 LongData;
            [FieldOffset(8)]
            public UInt64 Long64Data;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct OSVERSIONINFOEX
        {
            public uint dwOSVersionInfoSize;
            public uint dwMajorVersion;
            public uint dwMinorVersion;
            public uint dwBuildNumber;
            public uint dwPlatformId;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 128)]
            public string szCSDVersion;
            public ushort wServicePackMajor;
            public ushort wServicePackMinor;
            public ushort wSuiteMask;
            public byte wProductType;
            public byte wReserved;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct SYSTEM_INFO
        {
            public ushort wProcessorArchitecture;
            public ushort wReserved;
            public uint dwPageSize;
            public IntPtr lpMinimumApplicationAddress;
            public IntPtr lpMaximumApplicationAddress;
            public IntPtr dwActiveProcessorMask;
            public uint dwNumberOfProcessors;
            public uint dwProcessorType;
            public uint dwAllocationGranularity;
            public ushort wProcessorLevel;
            public ushort wProcessorRevision;
        }

        private enum PROCESSOR_ARCHITECTURE { UNDEFINED = 0, X86, X64 };
        private enum WINDOWS_VERSION { UNDEFINED = 0, Windows95, Windows98, Windows98SE, WindowsME, WindowsNT3, WindowsNT4, Windows2000, WindowsXP, Windows2003, WindowsVista, Windows2008, Windows7, Windows2008R2 }
        private struct OS_ENVIRONMENT
        {
            public PROCESSOR_ARCHITECTURE Arch;
            public WINDOWS_VERSION OsVersion;
            public bool IsWow64;
        }
        #endregion

        #region DataMember
        public static bool IsDriverReady = false;
        public static bool IsDeviceReady = false;
        private static IntPtr hDevice = Win32.InvalidHandleValue;
        #endregion

        #region Load and Unload
        protected static bool LoadDriver()
        {
            bool ret = false;
            IntPtr sc_handle = IntPtr.Zero;
            IntPtr sv_handle = IntPtr.Zero;
            OS_ENVIRONMENT OsEnv = GetEnvironment();
            IsDriverReady = false;

            sc_handle = Win32.OpenSCManager(null, null, Win32.SC_MANAGER_CREATE_SERVICE);
            if (sc_handle == IntPtr.Zero)
            {
                NuvHwAccCtrl.LogError(string.Format("LoadDriver(): OpenSCManager failed. error=0x{0:X}", Win32.GetLastError()));
                goto lab_ret;
            }

            string DriverPath = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + System.IO.Path.DirectorySeparatorChar + NUV_DRV_DIRECTORY + System.IO.Path.DirectorySeparatorChar + ((OsEnv.Arch == PROCESSOR_ARCHITECTURE.X64) ? NUV_DRV_BINARY_X64 : NUV_DRV_BINARY_X86);
            
            sv_handle = Win32.CreateService(sc_handle,
                                            NUV_DRV_SERVICE_NAME,
                                            NUV_DRV_DISPLAY_NAME,
                                            (uint)Win32.SERVICE_ACCESS.SERVICE_ALL_ACCESS,
                                            (uint)Win32.SERVICE_TYPE.SERVICE_KERNEL_DRIVER,
                                            (uint)Win32.SERVICE_START_TYPE.SERVICE_DEMAND_START,
                                            Win32.SERVICE_ERROR_NORMAL,
                                            DriverPath,
                                            null, 0, null, null, null);
            if (sv_handle == IntPtr.Zero)
            {
                uint err = Win32.GetLastError();
                if (err != Win32.ERROR_SERVICE_EXISTS)
                {
                    NuvHwAccCtrl.LogError(string.Format("LoadDriver(): CreateService failed. error=0x{0:X}", err));
                    NuvHwAccCtrl.LogInfo(string.Format("LoadDriver(): Driver Path: ") + DriverPath);
                    NuvHwAccCtrl.LogInfo(string.Format("LoadDriver(): Service Name: ") + NUV_DRV_SERVICE_NAME);
                    NuvHwAccCtrl.LogInfo(string.Format("LoadDriver(): Display Name: ") + NUV_DRV_DISPLAY_NAME);
                    goto lab_ret;
                }

                sv_handle = Win32.OpenService(sc_handle, NUV_DRV_SERVICE_NAME, (uint)Win32.SERVICE_ACCESS.SERVICE_ALL_ACCESS);
                if (sv_handle == IntPtr.Zero)
                {
                    NuvHwAccCtrl.LogError(string.Format("LoadDriver(): OpenService failed. error=0x{0:X}", Win32.GetLastError()));
                    goto lab_ret;
                }

                if (!Win32.ChangeServiceConfig(sv_handle,
                                               (uint)Win32.SERVICE_TYPE.SERVICE_KERNEL_DRIVER,
                                               (uint)Win32.SERVICE_START_TYPE.SERVICE_DEMAND_START,
                                               (uint)Win32.SERVICE_ERROR_NORMAL,
                                               DriverPath,
                                               null, 0, null, null, null,
                                               NUV_DRV_DISPLAY_NAME))
                {
                    NuvHwAccCtrl.LogError(string.Format("LoadDriver(): ChangeServiceConfig failed. error=0x{0:X}", Win32.GetLastError()));
                }
            }

            // Check if the service is already running.
            Win32.SERVICE_STATUS_PROCESS svStatus;
            IntPtr buf = IntPtr.Zero;
            int dwOldCheckPoint;
            int dwStartTickCount;
            int dwWaitTime;

            try
            {
                int size = 0;
                Win32.QueryServiceStatusEx(sv_handle, 0, buf, size, out size);
                buf = Marshal.AllocHGlobal(size);

                if (!Win32.QueryServiceStatusEx(sv_handle, 0, buf, size, out size))
                {
                    // throw new Win32Exception(Marshal.GetLastWin32Error());
                    // Log4Net.ReportMessageDebug(String::Format("LoadDriver(): QueryServiceStatusEx failed. error=0x{0:X}.", GetLastError()));
                }

                svStatus = (Win32.SERVICE_STATUS_PROCESS)Marshal.PtrToStructure(buf, typeof(Win32.SERVICE_STATUS_PROCESS));
                if (svStatus.CurrentState == Win32.SERVICE_STATE.SERVICE_STOP_PENDING)
	            {
                    dwStartTickCount = Win32.GetTickCount();
                    dwOldCheckPoint = svStatus.CheckPoint;
		            while (svStatus.CurrentState == Win32.SERVICE_STATE.SERVICE_STOP_PENDING)
		            {
			            // Do not wait longer than the wait hint. A good interval is one-tenth of the wait hint but not less than 1 second and not more than 10 seconds. 
			            dwWaitTime = svStatus.WaitHint / 10;
			            if (dwWaitTime < 1000)			dwWaitTime = 1000;
			            else if (dwWaitTime > 10000)	dwWaitTime = 10000;
                        System.Threading.Thread.Sleep(dwWaitTime);

                        if (!Win32.QueryServiceStatusEx(sv_handle, 0, buf, size, out size))
			            {
				            // Log4Net::ReportMessageDebug(String::Format("LoadDriver(): QueryServiceStatusEx failed. error=0x{0:X}.", GetLastError()));
				            goto lab_ret;
			            }
                        svStatus = (Win32.SERVICE_STATUS_PROCESS)Marshal.PtrToStructure(buf, typeof(Win32.SERVICE_STATUS_PROCESS));

			            if (svStatus.CheckPoint > dwOldCheckPoint)
			            {
				            // Continue to wait and check.
				            dwStartTickCount = Win32.GetTickCount();
				            dwOldCheckPoint = svStatus.CheckPoint;
			            }
			            else
			            {
                            if (Win32.GetTickCount() - dwStartTickCount > svStatus.WaitHint)
				            {
					            // Log4Net::ReportMessageDebug(String::Format("LoadDriver(): StopService Timeout."));
					            goto lab_ret;
				            }
			            }
		            }
	            }

	            if (svStatus.CurrentState == Win32.SERVICE_STATE.SERVICE_STOPPED)
	            {
		            // Attempt to start the service.
                    if (!Win32.StartService(sv_handle, 0, null))
                    {
                        uint err = Win32.GetLastError();
                        if (err != Win32.ERROR_SERVICE_ALREADY_RUNNING)
                        {
                            NuvHwAccCtrl.LogError(string.Format("LoadDriver(): StartService failed. error=0x{0:X}", err));
                            goto lab_ret;
                        }
                    }

    			    // Check the status until the service is no longer start pending. 
                    if (!Win32.QueryServiceStatusEx(sv_handle, 0, buf, size, out size))
		    	    {
                        // Log4Net::ReportMessageDebug(String::Format("LoadDriver(): QueryServiceStatusEx failed. error=0x{0:X}.", GetLastError()));
				        goto lab_ret;
    			    }
                    svStatus = (Win32.SERVICE_STATUS_PROCESS)Marshal.PtrToStructure(buf, typeof(Win32.SERVICE_STATUS_PROCESS));
                }
		            
                if (svStatus.CurrentState == Win32.SERVICE_STATE.SERVICE_START_PENDING)
                {
			        dwStartTickCount = Win32.GetTickCount();
			        dwOldCheckPoint = svStatus.CheckPoint;

                    while (svStatus.CurrentState == Win32.SERVICE_STATE.SERVICE_START_PENDING) 
			        {
                        // Do not wait longer than the wait hint. A good interval is one-tenth the wait hint, but no less than 1 second and no more than 10 seconds. 
				        dwWaitTime = svStatus.WaitHint / 10;
				        if (dwWaitTime < 1000)			dwWaitTime = 1000;
				        else if (dwWaitTime > 10000)	dwWaitTime = 10000;
                        System.Threading.Thread.Sleep(dwWaitTime);

                        if (!Win32.QueryServiceStatusEx(sv_handle, 0, buf, size, out size))
				        {
                            // Log4Net::ReportMessageDebug(String::Format("LoadDriver(): QueryServiceStatusEx failed. error=0x{0:X}.", GetLastError()));
					        goto lab_ret;
				        }
                        svStatus = (Win32.SERVICE_STATUS_PROCESS)Marshal.PtrToStructure(buf, typeof(Win32.SERVICE_STATUS_PROCESS));
 
				        if (svStatus.CheckPoint > dwOldCheckPoint)
				        {
                            // Continue to wait and check.
					        dwStartTickCount = Win32.GetTickCount();
					        dwOldCheckPoint = svStatus.CheckPoint;
				        }
				        else
				        {
                            if (Win32.GetTickCount()-dwStartTickCount > svStatus.WaitHint)
					        {
                                // No progress made within the wait hint.
						        break;
					        }
                        }
                    }
                }
	
                // Determine whether the service is running.
                if (svStatus.CurrentState == Win32.SERVICE_STATE.SERVICE_RUNNING) 
                {
		            // Log4Net::ReportMessageDebug(String::Format("LoadDriver(): Service started successfully."));
		            ret = true;
                }
                else 
                { 
		            // Log4Net::ReportMessageDebug(String::Format("LoadDriver(): Service not started.\n" +  
					//								 " Current State: %d\n" +  
					//								 " Exit Code: %d\n" +  
					//								 " Check Point: %d\n" +  
					//								 " Wait Hint: %d\n", svStatus.dwCurrentState, svStatus.dwWin32ExitCode, svStatus.dwCheckPoint, svStatus.dwWaitHint));
                } 
            }
            finally
            {
                if (!buf.Equals(IntPtr.Zero))
                {
                    Marshal.FreeHGlobal(buf);
                }
            }

        lab_ret:
            if (sv_handle != IntPtr.Zero) Win32.CloseServiceHandle(sv_handle);
            if (sc_handle != IntPtr.Zero) Win32.CloseServiceHandle(sc_handle);
            return ret;
        }

        protected static bool UnLoadDriver()
        {
            bool ret = false;
            IntPtr sc_handle = IntPtr.Zero;
            IntPtr sv_handle = IntPtr.Zero;

            IsDriverReady = false;
            sc_handle = Win32.OpenSCManager(null, null, Win32.SC_GENERIC_WRITE);
            if (sc_handle == IntPtr.Zero)
            {
                NuvHwAccCtrl.LogError(string.Format("UnloadDriver(): StartService failed. error=0x{0:X}", Win32.GetLastError()));
                goto lab_ret;
            }

            sv_handle = Win32.OpenService(sc_handle, NUV_DRV_SERVICE_NAME, (uint)Win32.SERVICE_ACCESS.SERVICE_ALL_ACCESS);
            if (sv_handle == IntPtr.Zero)
            {
                NuvHwAccCtrl.LogError(string.Format("UnloadDriver(): OpenService failed. error=0x{0:X}", Win32.GetLastError()));
                goto lab_ret;
            }

            Win32.SERVICE_STATUS stat = new Win32.SERVICE_STATUS();
            if (Win32.ControlService(sv_handle, Win32.SERVICE_CONTROL.STOP, ref stat) == false)
            {
                NuvHwAccCtrl.LogError(string.Format("UnloadDriver(): StopService failed. error=0x{0:X}", Win32.GetLastError()));
                //goto lab_ret;
            }
            Thread.Sleep(100);

            if (Win32.DeleteService(sv_handle) == false)
            {
                NuvHwAccCtrl.LogError(string.Format("UnloadDriver(): DeleteService failed. error=0x{0:X}", Win32.GetLastError()));
                goto lab_ret;
            }
            ret = true;

        lab_ret:
            if (sv_handle == IntPtr.Zero) Win32.CloseServiceHandle(sv_handle);
            if (sc_handle == IntPtr.Zero) Win32.CloseServiceHandle(sc_handle);
            return ret;
        }

        protected static bool InitDriver()
        {
            uint FILE_ATTRIBUTE_NORMAL = 0x80;

            IsDriverReady = false;
            ReleaseDriver();
            hDevice = Win32.CreateFile(NUV_DRV_DEVICE_PATH, Win32.GENERIC_READ | Win32.GENERIC_WRITE, 0, IntPtr.Zero, Win32.OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, IntPtr.Zero);
            if (hDevice == Win32.InvalidHandleValue)
            {
                NuvHwAccCtrl.LogError(string.Format("InitDriver(): Open driver failed. error=0x{0:X}. Try to load driver.", Win32.GetLastError()));
                hDevice = Win32.InvalidHandleValue;
                return false;
            }

            IsDriverReady = true;
            return true;
        }

        protected static void ReleaseDriver()
        {
            if (hDevice != Win32.InvalidHandleValue)
            {
                Win32.CloseHandle(hDevice);
                hDevice = Win32.InvalidHandleValue;
            }
        }
        #endregion Load and Unload

        #region Generic Port Registers Access
        protected static bool outportb(UInt16 Port, Byte Data)
        {
            if (IsDriverReady == false) return false;

            GENPORT_WRITE_INPUT WritePort = new GENPORT_WRITE_INPUT();
            WritePort.PortNumber = Port;
            WritePort.CharData = Data;

            IntPtr InData = Marshal.AllocHGlobal((Marshal.SizeOf(WritePort)));
            Marshal.StructureToPtr(WritePort, InData, false);
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PORT_UCHAR, InData, (uint)Marshal.SizeOf(WritePort), IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outportb failed. error=0x{0:X}", Win32.GetLastError()));
                Marshal.FreeHGlobal(InData);
                return false;
            }
            Marshal.FreeHGlobal(InData);
            return true;
        }

        protected static bool inportb(UInt16 Port, out Byte Data)
        {
            Data = 0xFF;

            if (IsDriverReady == false)
                return false;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(Port));
            Marshal.StructureToPtr(Port, InData, false);
            IntPtr OutData = Marshal.AllocHGlobal(Marshal.SizeOf(Data));
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PORT_UCHAR, InData, sizeof(UInt32), OutData, (uint)Marshal.SizeOf(Data), ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inportb failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(InData);
                Marshal.FreeHGlobal(OutData);
                return false;
            }
            Data = (Byte)Marshal.PtrToStructure(OutData, Data.GetType());
            Marshal.FreeHGlobal(InData);
            Marshal.FreeHGlobal(OutData);
            return true;
        }

        protected static bool outports(UInt16 Port, UInt16 Data)
        {
            if (IsDriverReady == false) return false;

            GENPORT_WRITE_INPUT WritePort = new GENPORT_WRITE_INPUT();
            WritePort.PortNumber = Port;
            WritePort.ShortData = Data;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(WritePort));
            Marshal.StructureToPtr(WritePort, InData, false);
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PORT_USHORT, InData, (uint)Marshal.SizeOf(WritePort), IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outports failed. error=0x{0:X}", Win32.GetLastError()));
                Marshal.FreeHGlobal(InData);
                return false;
            }
            Marshal.FreeHGlobal(InData);
            return true;
        }

        protected static bool inports(UInt16 Port, out UInt16 Data)
        {
            Data = 0xFFFF;

            if (IsDriverReady == false)
                return false;

            IntPtr InData = Marshal.AllocHGlobal((Marshal.SizeOf(Port)));
            Marshal.StructureToPtr(Port, InData, false);
            IntPtr OutData = Marshal.AllocHGlobal(Marshal.SizeOf(Data));
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PORT_USHORT, InData, sizeof(UInt32), OutData, (uint)Marshal.SizeOf(Data), ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inports failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(InData);
                Marshal.FreeHGlobal(OutData);
                return false;
            }
            Data = (UInt16)Marshal.PtrToStructure(OutData, Data.GetType());
            Marshal.FreeHGlobal(InData);
            Marshal.FreeHGlobal(OutData);
            return true;
        }

        protected static bool outportl(UInt16 Port, UInt32 Data)
        {
            if (IsDriverReady == false) return false;

            GENPORT_WRITE_INPUT WritePort = new GENPORT_WRITE_INPUT();
            WritePort.PortNumber = Port;
            WritePort.LongData = Data;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(WritePort));
            Marshal.StructureToPtr(WritePort, InData, false);
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PORT_ULONG, InData, (uint)Marshal.SizeOf(WritePort), IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outportl failed. error=0x{0:X}", Win32.GetLastError()));
                Marshal.FreeHGlobal(InData);
                return false;
            }
            Marshal.FreeHGlobal(InData);
            return true;
        }

        protected static bool inportl(UInt16 Port, out UInt32 Data)
        {
            Data = 0xFFFFFFFF;

            if (IsDriverReady == false)
                return false;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(Port));
            Marshal.StructureToPtr(Port, InData, false);
            IntPtr OutData = Marshal.AllocHGlobal(Marshal.SizeOf(Data));
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PORT_ULONG, InData, sizeof(UInt32), OutData, (uint)Marshal.SizeOf(Data), ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inportl failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(InData);
                Marshal.FreeHGlobal(OutData);
                return false;
            }
            Data = (UInt32)Marshal.PtrToStructure(OutData, Data.GetType());
            Marshal.FreeHGlobal(InData);
            Marshal.FreeHGlobal(OutData);
            return true;
        }

        protected static bool outportbuffer(UInt16 Port, Byte[] Buffer, UInt32 StartIndex, UInt32 Length)
        {
            if (IsDriverReady == false)
                return false;

            uint WriteLen = sizeof(UInt32) * 2 + sizeof(Byte) * Length;
            IntPtr WritePort = Marshal.AllocHGlobal((int)WriteLen);
            MarshalWriteUInt32(WritePort, 0, Port);
            MarshalWriteUInt32(WritePort, 4, Length);
            MarshalWriteByteAry(WritePort, 8, Buffer, StartIndex, Length);

            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PORT_BUFFER_UCHAR, WritePort, WriteLen, IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outportbuffer failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(WritePort);
                return false;
            }
            Marshal.FreeHGlobal(WritePort);
            return true;
        }

        protected static bool inportbuffer(UInt16 Port, ref Byte[] Buffer, UInt32 StartIndex, UInt32 Length)
        {
            for (uint i = 0; i < Length; i++)
                Buffer[StartIndex + i] = 0xFF;

            if (IsDriverReady == false)
                return false;

            uint ReadLen = (uint)(sizeof(UInt32) * 2);
            IntPtr ReadPort = Marshal.AllocHGlobal((int)ReadLen);
            MarshalWriteUInt32(ReadPort, 0, Port);
            MarshalWriteUInt32(ReadPort, 4, Length);
            uint OutDataLen = (uint)(sizeof(Byte) * Length);
            IntPtr OutData = Marshal.AllocHGlobal((int)OutDataLen);

            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PORT_BUFFER_UCHAR, ReadPort, ReadLen, OutData, OutDataLen, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inportbuffer failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(OutData);
                Marshal.FreeHGlobal(ReadPort);
                return false;
            }
            Marshal.Copy(OutData, Buffer, 0, (int)Length);
            Marshal.FreeHGlobal(OutData);
            Marshal.FreeHGlobal(ReadPort);
            return true;
        }
        #endregion Read and Write Port

        #region Memory Access
        protected static bool outmemb(UInt64 Address, Byte Data)
        {
            if (IsDriverReady == false) return false;

            GENMEM_WRITE_INPUT WriteMem = new GENMEM_WRITE_INPUT();
            WriteMem.MemAddr = Address;
            WriteMem.CharData = Data;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(WriteMem));
            Marshal.StructureToPtr(WriteMem, InData, false);
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PHY_MEM_UCHAR, InData, (uint)Marshal.SizeOf(WriteMem), IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outmemb failed. error=0x{0:X}", Win32.GetLastError()));
                Marshal.FreeHGlobal(InData);
                return false;
            }
            Marshal.FreeHGlobal(InData);
            return true;
        }

        protected static bool inmemb(UInt64 Address, out Byte Data)
        {
            Data = 0;

            if (IsDriverReady == false)
                return false;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(Address));
            Marshal.StructureToPtr(Address, InData, false);
            IntPtr OutData = Marshal.AllocHGlobal(Marshal.SizeOf(Data));
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PHY_MEM_UCHAR, InData, (uint)Marshal.SizeOf(Address), OutData, sizeof(byte), ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inmemb failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(InData);
                Marshal.FreeHGlobal(OutData);
                return false;
            }
            Data = (byte)Marshal.PtrToStructure(OutData, typeof(byte));
            Marshal.FreeHGlobal(InData);
            Marshal.FreeHGlobal(OutData);
            return true;
        }

        protected static bool outmems(UInt64 Address, UInt16 Data)
        {
            if (IsDriverReady == false) return false;

            GENMEM_WRITE_INPUT WriteMem = new GENMEM_WRITE_INPUT();
            WriteMem.MemAddr = Address;
            WriteMem.ShortData = Data;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(WriteMem));
            Marshal.StructureToPtr(WriteMem, InData, false);
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PHY_MEM_USHORT, InData, (uint)Marshal.SizeOf(WriteMem), IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outmems failed. error=0x{0:X}", Win32.GetLastError()));
                Marshal.FreeHGlobal(InData);
                return false;
            }
            Marshal.FreeHGlobal(InData);
            return true;
        }

        protected static bool inmems(UInt64 Address, out UInt16 Data)
        {
            Data = 0;

            if (IsDriverReady == false)
                return false;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(Address));
            Marshal.StructureToPtr(Address, InData, false);
            IntPtr OutData = Marshal.AllocHGlobal(Marshal.SizeOf(Data));
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PHY_MEM_USHORT, InData, (uint)Marshal.SizeOf(Address), OutData, (uint)Marshal.SizeOf(Data), ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inmems failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(InData);
                Marshal.FreeHGlobal(OutData);
                return false;
            }
            Data = (UInt16)Marshal.PtrToStructure(OutData, Data.GetType());
            Marshal.FreeHGlobal(InData);
            Marshal.FreeHGlobal(OutData);
            return true;
        }

        protected static bool outmeml(UInt64 Address, UInt32 Data)
        {
            if (IsDriverReady == false) return false;

            GENMEM_WRITE_INPUT WriteMem = new GENMEM_WRITE_INPUT();
            WriteMem.MemAddr = Address;
            WriteMem.LongData = Data;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(WriteMem));
            Marshal.StructureToPtr(WriteMem, InData, false);
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PHY_MEM_ULONG, InData, (uint)Marshal.SizeOf(WriteMem), IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outmeml failed. error=0x{0:X}", Win32.GetLastError()));
                Marshal.FreeHGlobal(InData);
                return false;
            }
            Marshal.FreeHGlobal(InData);
            return true;
        }

        protected static bool inmeml(UInt64 Address, out UInt32 Data)
        {
            Data = 0;

            if (IsDriverReady == false)
                return false;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(Address));
            Marshal.StructureToPtr(Address, InData, false);
            IntPtr OutData = Marshal.AllocHGlobal(Marshal.SizeOf(Data));
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PHY_MEM_ULONG, InData, (uint)Marshal.SizeOf(Address), OutData, (uint)Marshal.SizeOf(Data), ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inmeml failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(InData);
                Marshal.FreeHGlobal(OutData);
                return false;
            }
            Data = (UInt32)Marshal.PtrToStructure(OutData, Data.GetType());
            Marshal.FreeHGlobal(InData);
            Marshal.FreeHGlobal(OutData);
            return true;
        }

        protected static bool outmeml64(UInt64 Address, UInt64 Data)
        {
            if (IsDriverReady == false) return false;

            GENMEM_WRITE_INPUT WriteMem = new GENMEM_WRITE_INPUT();
            WriteMem.MemAddr = Address;
            WriteMem.Long64Data = Data;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(WriteMem));
            Marshal.StructureToPtr(WriteMem, InData, false);
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PHY_MEM_ULONG64, InData, (uint)Marshal.SizeOf(WriteMem), IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outmeml64 failed. error=0x{0:X}", Win32.GetLastError()));
                Marshal.FreeHGlobal(InData);
                return false;
            }
            Marshal.FreeHGlobal(InData);
            return true;
        }

        protected static bool inmeml64(UInt64 Address, out UInt64 Data)
        {
            Data = 0;

            if (IsDriverReady == false)
                return false;

            IntPtr InData = Marshal.AllocHGlobal(Marshal.SizeOf(Address));
            Marshal.StructureToPtr(Address, InData, false);
            IntPtr OutData = Marshal.AllocHGlobal(Marshal.SizeOf(Data));
            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PHY_MEM_ULONG64, InData, (uint)Marshal.SizeOf(Address), OutData, (uint)Marshal.SizeOf(Data), ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inmeml64 failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(InData);
                Marshal.FreeHGlobal(OutData);
                return false;
            }
            Data = (UInt64)Marshal.PtrToStructure(OutData, Data.GetType());
            Marshal.FreeHGlobal(InData);
            Marshal.FreeHGlobal(OutData);
            return true;
        }

        protected static bool outmembuffer(UInt64 Address, Byte[] Buffer, UInt32 StartIndex, UInt32 Length)
        {
            if (IsDriverReady == false)
                return false;

            uint WriteLen = sizeof(UInt64) + sizeof(UInt32) + sizeof(Byte) * Length;
            IntPtr WriteMem = Marshal.AllocHGlobal((int)WriteLen);
            MarshalWriteUInt64(WriteMem, 0, Address);
            MarshalWriteUInt32(WriteMem, 8, Length);
            MarshalWriteByteAry(WriteMem, 12, Buffer, StartIndex, Length);

            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_WRITE_PHY_MEM_BUFFER_UCHAR, WriteMem, WriteLen, IntPtr.Zero, 0, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("outmembuffer failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(WriteMem);
                return false;
            }
            Marshal.FreeHGlobal(WriteMem);
            return true;
        }

        protected static bool inmembuffer(UInt64 Address, ref Byte[] Buffer, UInt32 StartIndex, UInt32 Length)
        {
            for (uint i = 0; i < Length; i++)
                Buffer[StartIndex + i] = 0;

            if (IsDriverReady == false)
                return false;

            uint ReadLen = (uint)(sizeof(UInt64) + sizeof(UInt32));
            IntPtr ReadMem = Marshal.AllocHGlobal((int)ReadLen);
            MarshalWriteUInt64(ReadMem, 0, Address);
            MarshalWriteUInt32(ReadMem, 8, Length);
            uint OutDataLen = (uint)(sizeof(Byte) * Length);
            IntPtr OutData = Marshal.AllocHGlobal((int)OutDataLen);

            uint RetLength = 0;
            if (!Win32.DeviceIoControl(hDevice, IOCTL_GPD_READ_PHY_MEM_BUFFER_UCHAR, ReadMem, ReadLen, OutData, OutDataLen, ref RetLength, IntPtr.Zero))
            {
                Debug.WriteLine(string.Format("inmembuffer failed. error=0x{0:X}", Win32.GetLastError())); ;
                Marshal.FreeHGlobal(OutData);
                Marshal.FreeHGlobal(ReadMem);
                return false;
            }
            Marshal.Copy(OutData, Buffer, 0, (int)Length);
            Marshal.FreeHGlobal(OutData);
            Marshal.FreeHGlobal(ReadMem);
            return true;
        }
        #endregion

        #region Misc Method
        private static OS_ENVIRONMENT GetEnvironment()
        {
            OS_ENVIRONMENT OsEnv = new OS_ENVIRONMENT();
            OsEnv.Arch = PROCESSOR_ARCHITECTURE.UNDEFINED;
            OsEnv.OsVersion = WINDOWS_VERSION.UNDEFINED;
            OsEnv.IsWow64 = false;

            SYSTEM_INFO SysInfo;
            GetNativeSystemInfo(out SysInfo);
            switch (SysInfo.wProcessorArchitecture)
            {
                case PROCESSOR_ARCHITECTURE_INTEL:
                    OsEnv.Arch = PROCESSOR_ARCHITECTURE.X86;
                    break;
                case PROCESSOR_ARCHITECTURE_AMD64:
                    OsEnv.Arch = PROCESSOR_ARCHITECTURE.X64;
                    break;
            }

            IsWow64Process(Process.GetCurrentProcess().Handle, out OsEnv.IsWow64);

            OSVERSIONINFOEX osVersionInfo = new OSVERSIONINFOEX();
            osVersionInfo.dwOSVersionInfoSize = (uint)Marshal.SizeOf(typeof(OSVERSIONINFOEX));
            bool IsGetWin32Verson = GetVersionEx(ref osVersionInfo);

            if (osVersionInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
            {
                switch (osVersionInfo.dwMinorVersion)
                {
                    case 0:
                        OsEnv.OsVersion = WINDOWS_VERSION.Windows95;
                        break;
                    case 10:
                        if ((osVersionInfo.dwBuildNumber & 0xFFFF) == (uint)2222)
                            OsEnv.OsVersion = WINDOWS_VERSION.Windows98SE;
                        else
                            OsEnv.OsVersion = WINDOWS_VERSION.Windows98;
                        break;
                    case 90:
                        OsEnv.OsVersion = WINDOWS_VERSION.WindowsME;
                        break;
                    default:
                        break;
                }
            }
            else if (osVersionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT)
            {
                switch (osVersionInfo.dwMajorVersion)
                {
                    case 3:
                        OsEnv.OsVersion = WINDOWS_VERSION.WindowsNT3;
                        break;
                    case 4:
                        OsEnv.OsVersion = WINDOWS_VERSION.WindowsNT4;
                        break;
                    case 5:
                        if (osVersionInfo.dwMinorVersion == 0)
                            OsEnv.OsVersion = WINDOWS_VERSION.Windows2000;
                        else if (osVersionInfo.dwMinorVersion == 1)
                            OsEnv.OsVersion = WINDOWS_VERSION.WindowsXP;
                        else if (osVersionInfo.dwMinorVersion == 2)
                            OsEnv.OsVersion = WINDOWS_VERSION.Windows2003;
                        break;
                    case 6:
                        if (osVersionInfo.dwMinorVersion == 0)
                        {
                            if (osVersionInfo.wProductType == VER_NT_WORKSTATION)
                                OsEnv.OsVersion = WINDOWS_VERSION.WindowsVista;
                            else
                                OsEnv.OsVersion = WINDOWS_VERSION.Windows2008;
                        }
                        else if (osVersionInfo.dwMinorVersion == 1)
                        {
                            if (osVersionInfo.wProductType == VER_NT_WORKSTATION)
                                OsEnv.OsVersion = WINDOWS_VERSION.Windows7;
                            else
                                OsEnv.OsVersion = WINDOWS_VERSION.Windows2008R2;
                        }
                        break;
                    default:
                        break;
                }
            }

            Debug.WriteLine("Arch: " + OsEnv.Arch.ToString() + " OS: " + OsEnv.OsVersion.ToString() + " Wow64: " + OsEnv.IsWow64.ToString());
            return OsEnv;
        }

        private static void MarshalWriteUInt16(IntPtr ptr, Int32 ofs, UInt16 val)
        {
            byte[] ByteAry = BitConverter.GetBytes(val);
            for (int i = 0; i < ByteAry.Length; i++)
                Marshal.WriteByte(ptr, ofs + i, ByteAry[i]);
        }

        private static void MarshalWriteUInt32(IntPtr ptr, Int32 ofs, UInt32 val)
        {
            byte[] ByteAry = BitConverter.GetBytes(val);
            for (int i = 0; i < ByteAry.Length; i++)
                Marshal.WriteByte(ptr, ofs + i, ByteAry[i]);
        }

        private static void MarshalWriteUInt64(IntPtr ptr, Int32 ofs, UInt64 val)
        {
            byte[] ByteAry = BitConverter.GetBytes(val);
            for (int i = 0; i < ByteAry.Length; i++)
                Marshal.WriteByte(ptr, ofs + i, ByteAry[i]);
        }

        private static void MarshalWriteByteAry(IntPtr ptr, Int32 ofs, Byte[] array, UInt32 startIndex, UInt32 length)
        {
            for (int i = 0; i < length; i++)
                Marshal.WriteByte(ptr, ofs + i, array[startIndex + i]);
        }
        #endregion
    }
}
