﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Xml;
using System.Reflection;
using System.Threading;

using log4net;
using log4net.Config;

//using eSIO.Util;
//using eSIO.FunDpRefresher;

namespace NNuvHwAccCtrl
{
    public enum WBCHIP { NCT6116 };

    public class NuvHwAccCtrl : DriverManager
    {
        #region Const Definition
        private const string NUV_HW_LOCK = "NuvHwAccCtrl";
        private const string NUV_HM_LOCK = "NuvHealthManager"; //Don't modify this value. It is accross mullti-processes.  //HSY
        private const string NUV_MEM_LOCK = "NuvMemAccCtrl";   //Don't modify this value. It is accross mullti-processes.  //HSY

        private const byte LPC_LOGICAL_DEVICE_OFFSET = 0x07;
        private static WBCHIP[] ChipList = { WBCHIP.NCT6116};

        // public const byte CH0 = 0x00;
        // public const byte CH1 = 0x01;

        #endregion

        #region Structure Definition
        public struct NUVDRV_IOACCESS_CONFIG
        {
            public UInt16 Port;
            public byte[] Cmds;
        }
        #endregion

        #region Variable Definition
        public static Mutex HwAccessLock = new Mutex(false, NUV_HW_LOCK);
        public static Mutex HmAccessLock = new Mutex(false, NUV_HM_LOCK);
        public static Mutex MemAccessLock = new Mutex(false, NUV_MEM_LOCK);

        public static WBCHIP CHIP;
        private static UInt16 EFER, EFIR, EFDR;
        private static UInt16 SMBUSREG_BASE;
        private  static UInt16 PAGE0_HM, INDEX0_HM, DATA0_HM, PAGE1_HM, INDEX1_HM, DATA1_HM;

        private static byte bSrCfgRead;
        private static byte bSrCfgWrite;

        public static byte[] bData_WBuf = new byte[135];
        public static byte[] bData_RBuf = new byte[135];

        public static string IniPath;
        #endregion Variable Definition

        #region Load and Unload Method
        public static bool NuvHwCtrl_Load()
        {
            if (InitDriver() == false)
            {
                // only perform when driver and service has not been installed
                if (LoadDriver() == false) return false;
                if (InitDriver() == false) return false; 
            }

            LogInfo("NuvHwCtrl_Load OK.");
            return true;
        }

        public static void NuvHwCtrl_UnLoad()
        {
            LogInfo("Release Driver.");
            ReleaseDriver();
            UnLoadDriver();
        }
        #endregion Load and Unload Method

        #region Port Access
        public static byte InPortB(ushort Port)
        {
            byte Data;
            HmAccessLock.WaitOne();
            inportb(Port, out Data);
            HmAccessLock.ReleaseMutex();
            return Data;
        }

        public static void OutPortB(ushort Port, byte Data)
        {
            HmAccessLock.WaitOne();
            outportb(Port, Data);
            HmAccessLock.ReleaseMutex();
        }
/*
        public static ushort InPortS(ushort Port)
        {
            ushort Data;
            HmAccessLock.WaitOne();
            inports(Port, out Data);
            HmAccessLock.ReleaseMutex();
            return Data;
        }

        public static void OutPortS(ushort Port, ushort Data)
        {
            HmAccessLock.WaitOne();
            outports(Port, Data);
            HmAccessLock.ReleaseMutex();
        }
*/
        public static void InPortBuffer(ushort Port, ref byte[] Buffer, uint StartIndex, uint Length)
        {
            HmAccessLock.WaitOne();
            inportbuffer(Port, ref Buffer, StartIndex, Length);
            HmAccessLock.ReleaseMutex();
        }

        public static void OutPortBuffer(ushort Port, byte[] Buffer, uint StartIndex, uint Length)
        {
            HmAccessLock.WaitOne();
            outportbuffer(Port, Buffer, StartIndex, Length);
            HmAccessLock.ReleaseMutex();
        }
        #endregion

        #region Memory Access
/*
        public static byte InMemB(ulong Address)
        {
            byte Data;
            MemAccessLock.WaitOne();
            inmemb(Address, out Data);
            MemAccessLock.ReleaseMutex();
            return Data;
        }

        public static void OutMemB(ulong Address, byte Data)
        {
            MemAccessLock.WaitOne();
            outmemb(Address, Data);
            MemAccessLock.ReleaseMutex();
        }

        public static ushort InMemS(ulong Address)
        {
            ushort Data;
            MemAccessLock.WaitOne();
            inmems(Address, out Data);
            MemAccessLock.ReleaseMutex();
            return Data;
        }

        public static void OutMemS(ulong Address, ushort Data)
        {
            MemAccessLock.WaitOne();
            outmems(Address, Data);
            MemAccessLock.ReleaseMutex();
        }

        public static uint InMemL(ulong Address)
        {
            uint Data;
            MemAccessLock.WaitOne();
            inmeml(Address, out Data);
            MemAccessLock.ReleaseMutex();
            return Data;
        }

        public static void OutMemL(ulong Address, uint Data)
        {
            MemAccessLock.WaitOne();
            outmeml(Address, Data);
            MemAccessLock.ReleaseMutex();
        }

        public static ulong InMemL64(ulong Address)
        {
            ulong Data;
            MemAccessLock.WaitOne();
            inmeml64(Address, out Data);
            MemAccessLock.ReleaseMutex();
            return Data;
        }

        public static void OutMemL64(ulong Address, ulong Data)
        {
            MemAccessLock.WaitOne();
            outmeml64(Address, Data);
            MemAccessLock.ReleaseMutex();
        }

        public static void InMemBuffer(ulong Address, ref byte[] Buffer, uint StartIndex, uint Length)
        {
            MemAccessLock.WaitOne();
            inmembuffer(Address, ref Buffer, StartIndex, Length);
            MemAccessLock.ReleaseMutex();
        }

        public static void OutMemBuffer(ulong Address, byte[] Buffer, uint StartIndex, uint Length)
        {
            MemAccessLock.WaitOne();
            outmembuffer(Address, Buffer, StartIndex, Length);
            MemAccessLock.ReleaseMutex();
        }
 */ 
        #endregion

        #region Index / Data Access
        public static byte IdxInPortB(ushort IndexPort, ushort DataPort, byte Index, NUVDRV_IOACCESS_CONFIG[] OpenConfig, NUVDRV_IOACCESS_CONFIG[] CloseConfig)
        {
            byte Data = 0xFF;

            HmAccessLock.WaitOne();
            if (OpenConfig != null)
            {
                for (int i = 0; i < OpenConfig.Length; i++)
                {
                    if (OpenConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < OpenConfig[i].Cmds.Length; j++)
                        outportb(OpenConfig[i].Port, OpenConfig[i].Cmds[j]);
                }
            }

            outportb(IndexPort, Index);
            inportb(DataPort, out Data);

            if (CloseConfig != null)
            {
                for (int i = 0; i < CloseConfig.Length; i++)
                {
                    if (CloseConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < CloseConfig[i].Cmds.Length; j++)
                        outportb(CloseConfig[i].Port, CloseConfig[i].Cmds[j]);
                }
            }
            HmAccessLock.ReleaseMutex();
            return Data;
        }

        public static void IdxOutPortB(ushort IndexPort, ushort DataPort, byte Index, byte Data, NUVDRV_IOACCESS_CONFIG[] OpenConfig, NUVDRV_IOACCESS_CONFIG[] CloseConfig)
        {
            HmAccessLock.WaitOne();
            if (OpenConfig != null)
            {
                for (int i = 0; i < OpenConfig.Length; i++)
                {
                    if (OpenConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < OpenConfig[i].Cmds.Length; j++)
                        outportb(OpenConfig[i].Port, OpenConfig[i].Cmds[j]);
                }
            }

            outportb(IndexPort, Index);
            outportb(DataPort, Data);

            if (CloseConfig != null)
            {
                for (int i = 0; i < CloseConfig.Length; i++)
                {
                    if (CloseConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < CloseConfig[i].Cmds.Length; j++)
                        outportb(CloseConfig[i].Port, CloseConfig[i].Cmds[j]);
                }
            }
            HmAccessLock.ReleaseMutex();
        }
        #endregion

        #region Page / Index / Data Access
        public static bool PageInPortB(ushort PagePort, ushort IndexPort, ushort DataPort, byte Page, byte Index, out byte Data, bool LockMechanism)
        {
            byte RegVal;
            Data = 0xFF;

            HmAccessLock.WaitOne();
            if (LockMechanism == true)
            {
                // Check Page
                inportb(PagePort, out RegVal);
                if (RegVal != 0xFF)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }
                outportb(PagePort, Page);
                inportb(PagePort, out RegVal);
                if (RegVal != Page)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }

                // Check Index
                inportb(IndexPort, out RegVal);
                if (RegVal != 0xFF)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }
                outportb(IndexPort, Index);
                inportb(IndexPort, out RegVal);
                if (RegVal != Index)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }

                // Read Data
                inportb(DataPort, out Data);

                // Release Page Lock
                outportb(PagePort, 0xFF);
            }
            else
            {
                outportb(PagePort, Page);
                outportb(IndexPort, Index);
                inportb(DataPort, out Data);
            }
            HmAccessLock.ReleaseMutex();
            return true;
        }

        public static bool PageOutPortB(ushort PagePort, ushort IndexPort, ushort DataPort, byte Page, byte Index, byte Data, bool LockMechanism)
        {
            byte RegVal;

            HmAccessLock.WaitOne();
            if (LockMechanism == true)
            {
                // Check Page
                inportb(PagePort, out RegVal);
                if (RegVal != 0xFF)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }
                outportb(PagePort, Page);
                inportb(PagePort, out RegVal);
                if (RegVal != Page)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }

                // Check Index
                inportb(IndexPort, out RegVal);
                if (RegVal != 0xFF)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }
                outportb(IndexPort, Index);
                inportb(IndexPort, out RegVal);
                if (RegVal != Index)
                {
                    HmAccessLock.ReleaseMutex();
                    return false;
                }

                // Write Data
                outportb(DataPort, Data);

                // Release Page Lock
                outportb(PagePort, 0xFF);
            }
            else
            {
                outportb(PagePort, Page);
                outportb(IndexPort, Index);
                outportb(DataPort, Data);
            }
            HmAccessLock.ReleaseMutex();
            return true;
        }
        #endregion

        #region Hardware Monitor Access
        public static byte HmIdxInPortB(ushort IndexPort, ushort DataPort, byte BankIdx, byte Bank, byte Index)
        {
            byte Data = 0xFF;

            HmAccessLock.WaitOne();
            outportb(IndexPort, BankIdx);
            outportb(DataPort, Bank);
            outportb(IndexPort, Index);
            inportb(DataPort, out Data);
            HmAccessLock.ReleaseMutex();
            return Data;
        }

        public static void HmIdxOutPortB(ushort IndexPort, ushort DataPort, byte BankIdx, byte Bank, byte Index, byte Data)
        {
            HmAccessLock.WaitOne();
            outportb(IndexPort, BankIdx);
            outportb(DataPort, Bank);
            outportb(IndexPort, Index);
            outportb(DataPort, Data);
            HmAccessLock.ReleaseMutex();
        }
        #endregion

        #region LPC Configuration Space Access
        public static byte LpcInPortB(ushort IndexPort, ushort DataPort, byte LogicalDeviceNo, byte Index, NUVDRV_IOACCESS_CONFIG[] OpenConfig, NUVDRV_IOACCESS_CONFIG[] CloseConfig)
        {
            byte Data = 0xFF;

            HmAccessLock.WaitOne();
            if (OpenConfig != null)
            {
                for (int i = 0; i < OpenConfig.Length; i++)
                {
                    if (OpenConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < OpenConfig[i].Cmds.Length; j++)
                        outportb(OpenConfig[i].Port, OpenConfig[i].Cmds[j]);
                }
            }

            outportb(IndexPort, LPC_LOGICAL_DEVICE_OFFSET);
            outportb(DataPort, LogicalDeviceNo);
            outportb(IndexPort, Index);
            inportb(DataPort, out Data);

            if (CloseConfig != null)
            {
                for (int i = 0; i < CloseConfig.Length; i++)
                {
                    if (CloseConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < CloseConfig[i].Cmds.Length; j++)
                        outportb(CloseConfig[i].Port, CloseConfig[i].Cmds[j]);
                }
            }
            HmAccessLock.ReleaseMutex();
            return Data;
        }

        public static void LpcOutPortB(ushort IndexPort, ushort DataPort, byte LogicalDeviceNo, byte Index, byte Data, NUVDRV_IOACCESS_CONFIG[] OpenConfig, NUVDRV_IOACCESS_CONFIG[] CloseConfig)
        {
            HmAccessLock.WaitOne();
            if (OpenConfig != null)
            {
                for (int i = 0; i < OpenConfig.Length; i++)
                {
                    if (OpenConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < OpenConfig[i].Cmds.Length; j++)
                        outportb(OpenConfig[i].Port, OpenConfig[i].Cmds[j]);
                }
            }

            outportb(IndexPort, LPC_LOGICAL_DEVICE_OFFSET);
            outportb(DataPort, LogicalDeviceNo);
            outportb(IndexPort, Index);
            outportb(DataPort, Data);

            if (CloseConfig != null)
            {
                for (int i = 0; i < CloseConfig.Length; i++)
                {
                    if (CloseConfig[i].Cmds == null)
                        continue;
                    for (int j = 0; j < CloseConfig[i].Cmds.Length; j++)
                        outportb(CloseConfig[i].Port, CloseConfig[i].Cmds[j]);
                }
            }
            HmAccessLock.ReleaseMutex();
        }
        #endregion

        public static bool CheckSIO()
        {
            byte ucIdx, ucHi, ucLo;

            IsDeviceReady = false;
            // outportb(0x80, 0x37);

            for (ucIdx = 0; ucIdx < 2; ucIdx++)
            {
                if (ucIdx == 0)
                {
                    EFER = 0x4E;
                    LogDebug("Try 4E");
                }
                else
                {
                    EFER = 0x2E;
                    LogDebug("Try 2E");
                }

                EFIR = EFER;
                EFDR = (UInt16)(EFER + 1);
                outportb(EFER, 0xAA);
                outportb(EFER, 0x87);
                outportb(EFER, 0x87);

                outportb(EFIR, 0x20);
                inportb(EFDR, out ucHi);

                outportb(EFIR, 0x21);
                inportb(EFDR, out ucLo);

                LogInfo(string.Format("ChipID_Hi=0x{0:X}, ChipID_Lo=0x{1:X}.", ucHi, ucLo));

                if ((ucHi != 0xFF) || (ucLo != 0xFF))
                {
                    if ((ucHi == 0xD2) && ((ucLo & 0x80) == 0x80))
                    {
                        CHIP = WBCHIP.NCT6116;
                        IsDeviceReady = true;
                        return true;
                    }

                    //if (((ucHi == 0x55) && ((ucLo & 0xF0) == (0x38 & 0xF0))) ||
                    //    ((ucHi == 0xB2) && ((ucLo & 0xF0) == (0x71 & 0xF0))))
                    //{
                    //    CHIP = WBCHIP.eSIO;
                    //}
                    //else if ((ucHi == 0xC5) && ((ucLo & 0xF0) == (0x62 & 0xF0)))
                    //{
                    //    CHIP = WBCHIP.NCT6779;
                    //}
                    //else if ((ucHi == 0xC3) && ((ucLo & 0xF0) == (0x33 & 0xF0)))
                    //{
                    //    CHIP = WBCHIP.NCT6776;
                    //}
                    //else
                    //{
                    //    continue;
                    //}
                    /*
                                        // Get SMBus Port Resource setting
                                        if ((CHIP == WBCHIP.NCT6776) || (CHIP == WBCHIP.NCT6779))
                                        {
                                            outportb(EFIR, 0x07);
                                            outportb(EFDR, 0x0B);

                                            outportb(EFIR, 0x62);
                                            inportb(EFDR, out ucHi);

                                            outportb(EFIR, 0x63);
                                            inportb(EFDR, out ucLo);

                                            SMBUSREG_BASE = (UInt16)((UInt16)((UInt16)ucHi << 8) | (UInt16)ucLo);
                                        }
                                        else 
                                        if (CHIP == WBCHIP.eSIO)
                                        {
                                            outportb(EFIR, 0x07);
                                            outportb(EFDR, 0x0B);

                                            outportb(EFIR, 0x60);
                                            inportb(EFDR, out ucHi);

                                            outportb(EFIR, 0x61);
                                            inportb(EFDR, out ucLo);

                                            PAGE0_HM = (UInt16)((UInt16)((UInt16)ucHi << 8) | (UInt16)ucLo);
                                            INDEX0_HM = (UInt16)(PAGE0_HM + 1);
                                            DATA0_HM = (UInt16)(PAGE0_HM + 2);

                                            PAGE1_HM = (UInt16)(PAGE0_HM + 4);
                                            INDEX1_HM = (UInt16)(PAGE0_HM + 5);
                                            DATA1_HM = (UInt16)(PAGE0_HM + 6);
                                        }
                    */
                    CHIP = WBCHIP.NCT6116;
                    IsDeviceReady = true;
                    return true;
                }
            }

	        return false;
        }

        public static bool HM_ReadByte(byte Channel, byte Page, byte Index, ref byte retData)
        {
            byte ucTemp;
            byte ucTryCnt;

            ucTryCnt = 0;
            while (true)
            {
                if (Channel == 0)
                {
                    outportb(PAGE0_HM, Page);
                    inportb(PAGE0_HM, out ucTemp);
                }
                else
                {
                    outportb(PAGE1_HM, Page);
                    inportb(PAGE1_HM, out ucTemp);
                }

                if (ucTemp == Page) break;
                if (ucTryCnt == 0xFF) return false;
                ucTryCnt = (byte)(ucTryCnt + 1);
            }

            ucTryCnt = 0;
            while (true)
            {
                if (Channel == 0)
                {
                    outportb(INDEX0_HM, Index);
                    inportb(INDEX0_HM, out ucTemp);
                }
                else
                {
                    outportb(INDEX1_HM, Index);
                    inportb(INDEX1_HM, out ucTemp);
                }

                if (ucTemp == Index) break;
                if (ucTryCnt == 0xFF) return false;
                ucTryCnt = (byte)(ucTryCnt + 1);
            }

            if (Channel == 0)   inportb(DATA0_HM, out ucTemp);
            else                inportb(DATA1_HM, out ucTemp);
            retData = ucTemp;

            if (Channel == 0)   outportb(PAGE0_HM, 0xFF);
            else                outportb(PAGE1_HM, 0xFF);
            return true;
        }

        public static bool HM_WriteByte(byte Channel, byte Page, byte Index, byte Data)
        {
            byte ucTemp;
            byte ucTryCnt;

            ucTryCnt = 0;
            while (true)
            {
                if (Channel == 0)
                {
                    outportb(PAGE0_HM, Page);
                    inportb(PAGE0_HM, out ucTemp);
                }
                else
                {
                    outportb(PAGE1_HM, Page);
                    inportb(PAGE1_HM, out ucTemp);
                }

                if (ucTemp == Page) break;
                if (ucTryCnt == 0xFF) return false;
                ucTryCnt = (byte)(ucTryCnt + 1);
            }

            ucTryCnt = 0;
            while (true)
            {
                if (Channel == 0)
                {
                    outportb(INDEX0_HM, Index);
                    inportb(INDEX0_HM, out ucTemp);
                }
                else
                {
                    outportb(INDEX1_HM, Index);
                    inportb(INDEX1_HM, out ucTemp);
                }

                if (ucTemp == Index) break;
                if (ucTryCnt == 0xFF) return false;
                ucTryCnt = (byte)(ucTryCnt + 1);
            }

            if (Channel == 0)   outportb(DATA0_HM, Data);
            else                outportb(DATA1_HM, Data);

            if (Channel == 0)   outportb(PAGE0_HM, 0xFF);
            else                outportb(PAGE1_HM, 0xFF);
            return true;
        }

        #region Gpio access
        public static bool _EnableGPIO(int PortId, int BitId)
        {
            byte Data = 0xFF;

            HwAccessLock.WaitOne();

            if (CHIP == WBCHIP.NCT6116)
            {
                switch (PortId)
                {
                    case 0:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-function Pin
                        // 1Dh, Bit[4] = 0b
                        outportb(EFIR, 0x1D);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xEF);
                        outportb(EFDR, Data);

                        // 1Ch, Bit[3:2] = 00b
                        outportb(EFIR, 0x1C);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xF3);
                        outportb(EFDR, Data);

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE0);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    case 1:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-function Pin
                        // 1Ch, Bit[7] = 0b
                        outportb(EFIR, 0x1C);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0x7F);
                        outportb(EFDR, Data);

                        // 1Dh, Bit[1:0] = 00b
                        outportb(EFIR, 0x1D);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xFC);
                        outportb(EFDR, Data);

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE1);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    case 2:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-function Pin
                        // 2Bh, Bit[7] = 0b
                        outportb(EFIR, 0x2B);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0x7F);
                        outportb(EFDR, Data);

                        // 1Ch, Bit[1] = 0b
                        outportb(EFIR, 0x1C);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xFD);
                        outportb(EFDR, Data);

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE2);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    case 3:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-function Pin
                        // 1Ah, Bit[5:4] = 00b
                        outportb(EFIR, 0x1A);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xCF);
                        outportb(EFDR, Data);

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE3);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    case 4:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-function Pin
                        // 1Ah, Bit[5:4] = 00b
                        outportb(EFIR, 0x1A);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xCF);
                        outportb(EFDR, Data);

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE4);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    case 5:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-Function Pin
                        if (BitId == 0)
                        {
                            // (KBC_EN) 2Fh, Bit[4] = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xEF);
                            outportb(EFDR, Data);

                            // 1Ah, Bit[3] = 1b
                            outportb(EFIR, 0x1A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x08);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 1)
                        {
                            // KBC_EN = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xEF);
                            outportb(EFDR, Data);

                            // 1Ah, Bit[3] = 1b
                            outportb(EFIR, 0x1A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x08);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 2)
                        {
                            // KBC_EN = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xEF);
                            outportb(EFDR, Data);

                            // 1Ah, Bit[2] = 0b
                            outportb(EFIR, 0x1A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 3)
                        {
                            // KBC_EN = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xEF);
                            outportb(EFDR, Data);

                            // 1Ah, Bit[2] = 0b
                            outportb(EFIR, 0x1A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 4)
                        {
                            // 1Ah, Bit[1] = 1b
                            outportb(EFIR, 0x1A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x02);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 5)
                        {
                            // 1Ah, Bit[0] = 1b
                            outportb(EFIR, 0x1A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x01);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 6)
                        {
                            // 1Bh, Bit[7] = 1b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x80);
                            outportb(EFDR, Data);

                            // LDN_E, E6h, Bit[6] = 0b
                            outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                            outportb(EFDR, 0x0E);

                            outportb(EFIR, 0xE6);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xBF);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 7)
                        {
                            // 1Bh, Bit[6] = 1b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x40);
                            outportb(EFDR, Data);
                        }

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE5);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    case 6:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-function Pin
                        if (BitId == 0)
                        {
                            // 1Bh, Bit[5] = 1b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x20);
                            outportb(EFDR, Data);

                        }
                        else if (BitId == 1)
                        {
                            // 1Bh, Bit[4] = 1b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x10);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 2)
                        {
                            // 1Bh, Bit[3] = 1b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x08);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 3)
                        {
                            // 1Bh, Bit[1:0] = 00b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFC);
                            outportb(EFDR, Data);

                            // 2Ah, Bit[1] = 0b
                            outportb(EFIR, 0x2A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFD);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 4)
                        {
                            // 1Bh, Bit[1:0] = 00b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFC);
                            outportb(EFDR, Data);

                            // 2Ah, Bit[1] = 0b
                            outportb(EFIR, 0x2A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFD);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 5)
                        {
                            // DSW_EN = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);

                            // 1Ch, Bit[6] = 1b
                            outportb(EFIR, 0x1C);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x40);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 6)
                        {
                            // DSW_EN = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);

                            // 1Ch, Bit[5] = 1b
                            outportb(EFIR, 0x1C);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x20);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 7)
                        {
                            // 27h, Bit[2] = 0b
                            outportb(EFIR, 0x27);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);

                            // 1Ch, Bit[4] = 1b
                            outportb(EFIR, 0x1C);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x10);
                            outportb(EFDR, Data);
                        }

                        // Multi-function Select
                        if (BitId == 7)
                        {
                            outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                            outportb(EFDR, 0x08);

                            outportb(EFIR, 0xE7);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0x1F);
                            outportb(EFDR, Data);
                        }
                        else
                        {
                            outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                            outportb(EFDR, 0x08);

                            outportb(EFIR, 0xE6);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & ~(byte)(1 << BitId));
                            outportb(EFDR, Data);
                        }
                        break;

                    case 7:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data | ((byte)1 << PortId));
                        outportb(EFDR, Data);

                        // Multi-function Pin
                        // (PORT80_Strp) 2Fh, Bit[1] = 0b
                        outportb(EFIR, 0x2F);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xFD);
                        outportb(EFDR, Data);

                        // 2Bh, Bit[5] = 0b
                        outportb(EFIR, 0x2B);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & 0xDF);
                        outportb(EFDR, Data);

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE8);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    case 8:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x09);
                        outportb(EFIR, 0x30);
                        inportb(EFDR, out Data);

                        Data = (byte)(Data | ((byte)1 << (PortId - 8)));
                        outportb(EFDR, Data);

                        if (BitId == 0)
                        {
                            // 2Bh, Bit[4] = 1b
                            outportb(EFIR, 0x2B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x10);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 1)
                        {
                            // 2Bh, Bit[4] = 1b
                            outportb(EFIR, 0x2B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x10);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 2)
                        {
                            // 2Bh, Bit[3] = 0b
                            outportb(EFIR, 0x2B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xF7);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 3)
                        {
                            // DSW_EN = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 4)
                        {
                            // (WDT_EN) 2Dh, Bit[0] = 0b
                            outportb(EFIR, 0x2D);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFE);
                            outportb(EFDR, Data);

                            // 2Ah, Bit[3] = 1b
                            outportb(EFIR, 0x2A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x08);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 5)
                        {
                            // LDN_E, E6h, Bit[7] = 1b
                            outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                            outportb(EFDR, 0x0E);

                            outportb(EFIR, 0xE6);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x80);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 6)
                        {
                            // DSW_EN = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);

                            // 1Ah, Bit[7] = 0b
                            outportb(EFIR, 0x1A);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0x7F);
                            outportb(EFDR, Data);

                            // 1Bh, Bit[2] = 0b
                            outportb(EFIR, 0x1B);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);

                            // 1Ch, Bit[7] = 0b
                            outportb(EFIR, 0x1C);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0x7F);
                            outportb(EFDR, Data);
                        }
                        else if (BitId == 7)
                        {
                            // (DSW_EN) 2Fh, Bit[2] = 0b
                            outportb(EFIR, 0x2F);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data & 0xFB);
                            outportb(EFDR, Data);

                            // 1Dh, Bit[3] = 1b
                            outportb(EFIR, 0x1D);
                            inportb(EFDR, out Data);
                            Data = (byte)(Data | 0x08);
                            outportb(EFDR, Data);
                        }

                        // Multi-function Select
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x08);

                        outportb(EFIR, 0xE9);
                        inportb(EFDR, out Data);
                        Data = (byte)(Data & ~(byte)(1 << BitId));
                        outportb(EFDR, Data);
                        break;

                    default:
                        break;
                }
            }

            HwAccessLock.ReleaseMutex();
            return true;
        }

        public static byte _GetGpioMode(int PortId, int BitId)
        {
            byte Data = 0xFF;

            if (CHIP == WBCHIP.NCT6116)
            {
                switch (PortId)
                {
                    case 0:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE0);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 1:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE4);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 2:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE8);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 3:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xEC);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 4:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF0);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 5:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF4);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 6:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF8);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 7:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xFC);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;

                    case 8:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x09);

                        outportb(EFIR, 0xF0);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else                                  Data = 0x01;
                        break;
                }
            }

            return Data;
        }

        public static bool _SetGpioMode(int PortId, int BitId, byte Direction)
        {
            byte Data = 0xFF;

            if (CHIP == WBCHIP.NCT6116)
            {
                switch (PortId)
                {
                    case 0:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE0);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else                Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE0);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 1:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE4);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE1);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 2:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE8);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE2);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 3:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xEC);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE3);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 4:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF0);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE4);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 5:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF4);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE5);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 6:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF8);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE6);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 7:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xFC);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE7);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    case 8:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x09);

                        outportb(EFIR, 0xF0);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        else Data = (byte)(Data | (byte)(1 << BitId));
                        outportb(EFDR, Data);

                        // PP in output mode
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x0F);

                        outportb(EFIR, 0xE8);
                        inportb(EFDR, out Data);

                        if (Direction == 0) Data = (byte)(Data & ~((byte)(1 << BitId)));
                        outportb(EFDR, Data);
                        break;

                    default:
                        break;

                }
            }

            return true;
        }

        public static byte _GetGpioData(int PortId, int BitId)
        {
            byte Data = 0xFF;

            if (CHIP == WBCHIP.NCT6116)
            {
                switch (PortId)
                {
                    case 0:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE1);
                        inportb(EFDR, out Data);
                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 1:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE5);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 2:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE9);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 3:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xED);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 4:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF1);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 5:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF5);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 6:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF9);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 7:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xFD);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;

                    case 8:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x09);

                        outportb(EFIR, 0xF1);
                        inportb(EFDR, out Data);

                        if ((Data & (byte)(1 << BitId)) == 0) Data = 0x00;
                        else Data = 0x01;
                        break;
                }
            }

            LogDebug(string.Format("2, GP{0:d}{1:d}, data=0x{0:X}", PortId, BitId, Data));
            return Data;
        }

        public static bool _SetGpioData(int PortId, int BitId, byte Data)
        {
            byte bData = 0xFF;

            if (CHIP == WBCHIP.NCT6116)
            {
                switch (PortId)
                {
                    case 0:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE1);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 1:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE5);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 2:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xE9);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 3:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xED);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 4:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF1);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 5:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF5);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 6:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xF9);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 7:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x07);

                        outportb(EFIR, 0xFD);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    case 8:
                        outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                        outportb(EFDR, 0x09);

                        outportb(EFIR, 0xF1);
                        inportb(EFDR, out bData);

                        if (Data == 0) bData = (byte)(bData & ~((byte)(1 << BitId)));
                        else bData = (byte)(bData | (byte)(1 << BitId));
                        outportb(EFDR, bData);
                        break;

                    default:
                        break;
                }
            }

            return true;
        }
        #endregion

        #region WDTO
        public static bool _EnableWDTO(int val, int unit)
        {
            byte bData = 0xFF;

            if (CHIP == WBCHIP.NCT6116)
            {
                // 2Ah, Bit[3] = 0b
                outportb(EFIR, 0x2A);
                inportb(EFDR, out bData);
                bData = (byte)(bData & 0xF7);
                outportb(EFDR, bData);

                // (WDT_EN) 2Dh, Bit[0] = 0b
                outportb(EFIR, 0x2D);
                inportb(EFDR, out bData);
                bData = (byte)(bData & 0xFE);
                outportb(EFDR, bData);

                // WDT1
                outportb(EFIR, LPC_LOGICAL_DEVICE_OFFSET);
                outportb(EFDR, 0x08);

                outportb(EFIR, 0x30);
                inportb(EFDR, out bData);
                bData = (byte)(bData | 0x01);
                outportb(EFDR, bData);

                outportb(EFIR, 0xF0);
                inportb(EFDR, out bData);
                bData = (byte)(bData & 0xE0);
                // bData = (byte)(bData | 0x02); // use KBRST# to reset system

                if (unit == 1) bData = (byte)(bData | 0x08); // unit = min
                outportb(EFDR, bData);

                outportb(EFIR, 0xF1);
                outportb(EFDR, (byte)val);

                outportb(EFIR, 0xF2);
                inportb(EFDR, out bData);
                bData = (byte)(bData & 0xE0);
                if (unit == 1) bData = (byte)(bData & 0xEF); // clear status bit
                outportb(EFDR, bData);
            }

            return true;
        }
        #endregion

        #region ini access
        public static void IniWriteValue(string Section, string Key, string Value)
        {
            Win32.WritePrivateProfileString(Section, Key, Value, IniPath);
        }

        public static string IniReadValue(string Section, string Key)
        {
            StringBuilder temp = new StringBuilder(255);
            int i = Win32.GetPrivateProfileString(Section, Key, "", temp, 255, IniPath);
            return temp.ToString();
        }
        #endregion

        #region Log4Net
        private static readonly ILog Log = LogManager.GetLogger("NuvHwAccCtrl");
        private static string installPath = string.Empty;

        public static void LogInfo(string LogStr)
        {
            Log.Info(LogStr);
        }

        public static void LogError(string LogStr)
        {
            Debug.WriteLine(LogStr);
            Log.Error(LogStr);
        }

        public static void LogDebug(string LogStr)
        {
            Log.Debug(LogStr);
        }
        #endregion
    }
}
